#!/usr/bin/env python
"""Generates iCalendar files for BBC programmes."""
from __future__ import with_statement
import logging
import os
import sys
import time
import urllib2
import vobject
try:
    import json
# Fall-back for Python 2.5 and earlier
except ImportError:
    import simplejson as json
from functools import partial
from datetime import date, timedelta, tzinfo
from dateutil.parser import parse


logging.basicConfig(level=logging.INFO, format="%(message)s")


# I don't see a way to get a list of the service ids. So here are the service
# names and the regional variations for those that have them. Some channels
# have 3 weeks of future schedules. Some have 7 days.
bbc1_regions = (
    ('cambridge', 'Cambridgeshire'),
    ('channel_islands', 'Channel Islands'),
    ('east', 'East'),
    ('east_midlands', 'East Midlands'),
    ('london', 'London'),
    ('north_east', 'North East & Cumbria'),
    ('north_west', 'North West'),
    ('ni', 'Northern Ireland'),
    ('oxford', 'Oxfordshire'),
    ('scotland', 'Scotland'),
    ('south', 'South'),
    ('south_east', 'South East'),
    ('south_west', 'South West'),
    ('wales', 'Wales'),
    ('west', 'West'),
    ('west_midlands', 'West Midlands'),
    ('east_yorkshire', 'Yorks & Lincs'),
    ('yorkshire', 'Yorkshire'),
)


bbc2_regions = (
    ('england', 'England'),
    ('ni', 'Northern Ireland'),
    ('ni_analogue', 'Northern Ireland (Analogue)'),
    ('scotland', 'Scotland'),
    ('wales', 'Wales'),
    ('wales_analogue', 'Wales (Analogue)'),
)

radio1_regions = (
    ('england', 'England'),
    ('northernireland', 'Northern Ireland'),    # Note difference to TV region!
    ('scotland', 'Scotland'),
    ('wales', 'Wales'),
)

radio4_regions = (
    ('fm', 'FM'),
    ('lw', 'Long wave'),
)

services = (
    # Television
    ('bbcone', 'BBC 1', bbc1_regions),
    ('bbctwo', 'BBC 2', bbc2_regions),
    ('bbcthree', 'BBC 3'),   # No regional variation
    ('bbcfour', 'BBC 4'),
    ('bbchd', 'BBC HD'),
    ('cbeebies', 'CBeebies'),
    ('cbbc', 'CBBC'),
    ('bbcnews', 'BBC News'),
    ('parliament', 'Parliament'),   # No Funkadelic?
    ('radio1', 'Radio 1', radio1_regions),
    ('1xtra', '1Xtra'),
    ('radio2', 'Radio 2'),
    ('radio3', 'Radio 3'),
    ('radio4', 'Radio 4', radio4_regions),
    ('5live', '5 Live'),
    ('5livesportsextra', '5 Live Sports Extra'),
    ('6music', '6music'),
    ('radio7', 'Radio 7'),
    ('asiannetwork', 'Asian Network'),
    ('worldservice', 'World Service'),
)

# Templates for fetching the data
BBC_URL = 'http://www.bbc.co.uk'
NATIONAL_TEMPLATE = BBC_URL + "/%(service)s/programmes/schedules"
REGIONAL_TEMPLATE = BBC_URL + "/%(service)s/programmes/schedules/%(outlet)s"
# Used to build URLs to a programme's detail page
PID_TEMPLATE = BBC_URL + '/programmes/%(pid)s'

FETCH_DELAY = 3 # Delay in seconds between fetching URLs
REGION_SEPARATOR = '/'
ZERO = timedelta(0)


class UTC(tzinfo):
    """UTC"""
    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return ZERO


utc = UTC()


def all_service_info():
    """Returns a list of all known service identifiers. Each member is a couple
    of (service_id, service_name).
    """
    def _func():
        for service in services:
            if len(service) == 3:
                for region_id, region_name in service[2]:
                    service_id = service[0] + REGION_SEPARATOR + region_id
                    yield service_id, service[1] + ' ' + region_name
            else:
                yield service[0], service[1]
    return list(_func())
    

# Convenient list of all the service ids, urls and pretty names
services_info = all_service_info()


def polite_urlopen(url, **headers):
    """Wrapper for urllib2.urlopen that adds a custom User-Agent."""
    logging.debug('Open URL %s', url)
    req = urllib2.Request(url)
    if 'User-Agent'.lower() not in [k.lower() for k in headers.keys()]:
        headers['User-Agent'] = 'guide/0.1 http://reliablybroken.com'
    for key in headers:
        req.add_header(key, headers[key])
    opener = urllib2.build_opener()
    return opener.open(req)

    
def fixup_programme_urls(seq):
    """Re-writes the programme ID to an absolute URL for each event tuple in a
    sequence. Appends the URL to the event description. Returns a generator.
    """
    for event in seq:
        pid = event[4]
        url = PID_TEMPLATE % {'pid': pid}
        desc = event[3] + '\n\n' + url
        # Taking special care in case event is not actually a tuple
        yield tuple(event[:3]) + (desc, url)

    
def write_service_names(filename):
    """Writes a formatted list of service names and identifiers to an open file.
    """
    from operator import itemgetter
    
    name_width = max(len(s[1]) for s in services_info) + 1   # Width of first column
    filename.write('%s %s\n' % ('Service'.ljust(name_width), 'Identifier'))
    filename.write('%s %s\n' % ('-------'.ljust(name_width), '----------'))
    for service in sorted(services_info, key=itemgetter(1)):
        svc_id, svc_name = service[0], service[1]
        filename.write('%s %s\n' % (svc_name.ljust(name_width), svc_id))
        

def write_icalendar(filename, events, name, desc):
    """Creates a new iCalendar with the events and writes it to filename."""
    ical = vobject.iCalendar()
    ical.add('x-wr-calname').value = name
    ical.add('x-wr-caldesc').value = desc
    for ev in events:
        ical.add(ev)
    
    logging.info('Writing ics to %r', filename)
    filename.write(ical.serialize())
    
        
def make_calendars(services_info, dest):
    day_before_yesterday = date.today() - timedelta(days=2)
    
    for sid, sname in services_info:
        desc = 'Programme schedule for %s' % sname
        filename = os.path.join(dest, sid + '.ics')
        
        # TODO: save time and money by using old ics data
        #old_events = events_from_icalendar(open(filename))
        
        logging.info('Fetching data for %s', sid)
        new_events = set()
        for x in range(10):
            url = url_for_date(sid, day_before_yesterday + timedelta(days=x))
            try:
                new_events.update(events_from_json(polite_urlopen(url)))
            except urllib2.URLError:
                # 404 response or other error.
                pass

        new_events = sorted(set(new_events))
        # Make proper URLs from programme IDs
        new_events = fixup_programme_urls(new_events)
        
        all_events = create_vevents(new_events)
        with open(filename, 'wb') as out:
            write_icalendar(out, all_events, sname, desc)
        time.sleep(FETCH_DELAY)


def create_vevents(infos):
    """Returns a generator that produces VEvent objects.
    
    infos must be a sequence where each item is a 5-tuple of (start-dt,
    duration, title, description, url).
    """
    for dtstart, duration, summary, description, url in infos:
        event = vobject.newFromBehavior('vevent')
        event.add('dtstart').value = dtstart
        event.add('duration').value = duration
        event.add('summary').value = summary
        event.add('description').value = description
        event.add('url').value = url
        yield event


def url_template_for_service(service):
    """Returns the correct daily URL template for the given service."""
    if REGION_SEPARATOR in service:
        vars = dict(zip(('service', 'outlet'), service.split(REGION_SEPARATOR)))
        template = REGIONAL_TEMPLATE
    else:
        vars = dict(service=service)
        template = NATIONAL_TEMPLATE
    daily_template = (template % vars) + '/%(year)d/%(month)d/%(day)d.json'
    return daily_template


def url_for_date(service, dt):
    """Returns an URL string for the service on the date with a JSON ext."""
    template = url_template_for_service(service)
    return template % {'year': dt.year, 'month': dt.month, 'day': dt.day}
    
    
def events_from_json(filename):
    """Returns a generator of event information.
    
    Each item is a tuple of (start-dt, duration-delta, title, description, pid).
    
    The first item should be a datetime object so that we can sort the sequence
    later. Note that the last item is an opaque programme ID, not an URL.
    """
    data = json.load(filename)
    for broadcast in data['schedule']['day']['broadcasts']:
        yield (
            # vobject ignores dateutil's tzinfo objects so convert to UTC first.
            parse(broadcast['start']).astimezone(utc),
            timedelta(seconds=broadcast['duration']),
            # BBC's json puts numeric strings as integers
            unicode(broadcast['programme']['display_titles']['title']),
            unicode(broadcast['programme']['short_synopsis']),
            unicode(broadcast['programme']['pid']),
        )

    
def events_from_icalendar(filename):
    """Returns a generator of calendar event information.
    
    Each item is a tuple of (start-dt, duration-delta, title, description, url).
    """
    for root in vobject.readComponents(filename.read()):
        for e in root.getChildren():
            if e.name == 'VEVENT':
                yield (
                    e.dtstart.value,
                    e.duration.value,
                    e.summary.value,
                    e.description.value,
                    e.url.value,
                )
    
    
def usage(argv):
    sys.stderr.write("""
Generate iCalendar files for BBC schedules.

Usage:  %(prog)s -d <directory>

Options:
 -d, --directory  Directory for generated iCalendar files
 -s, --service    Generate iCalendar for a specific service (default is all services)
 -q, --quiet      Suppress most messages
 -l, --list       List known channels
 -h, --help       Show this help message
""" % {'prog':argv[0]})


def main(argv):
    import getopt
    
    flags = 'd:s:qlh'
    lflags = ('directory=', 'service=', 'quiet', 'list', 'help', 'debug')
    try:
        opts, args = getopt.getopt(argv[1:], flags, lflags)
    except getopt.GetoptError, err:
        sys.stderr.write(str(err))
        usage(argv)
        sys.exit(2)
    
    dest_dir = '.'
    services = set()
    list_services = False
    logger = logging.getLogger()
    
    for k, v in opts:
        if k in ('-d', '--directory'):
            dest_dir = v
        elif k in ('-s', '--service'):
            services.add(v)
        elif k in ('-q', '--quiet'):
            logger.setLevel(logging.WARNING)
        elif k in ('--debug',):
            logger.setLevel(logging.DEBUG)
        elif k in ('-h', '--help'):
            usage(argv)
            sys.exit(1)
        elif k in ('-l', '--list'):
            list_services = True
    
    if list_services:
        write_service_names(sys.stdout)
        sys.exit(0)
    else:
        infos = services_info
        if services:
            infos = [row for row in infos if row[0] in services]
        make_calendars(infos, dest_dir)
            

if __name__ == "__main__":
    try:
        main(sys.argv)
    except KeyboardInterrupt:
        sys.exit(1)
    
